from __future__ import annotations

from dataclasses import asdict, dataclass, field
from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, Iterator, List

from funcy import join
from typing_extensions import Literal, TypedDict

from .fields import FIELDS_MAP, get_val
from .generic import flexitable
from .utils import (
    border_panel,
    format_with_color,
    human_dt,
    new_tree,
    predictably_random_color,
    wrap,
)

if TYPE_CHECKING:
    from rich.panel import Panel
    from rich.tree import Tree

SKIP_HEADERS = {
    "priority",
    "recur",
    "uuid",
    "reviewed",
    "modified",
    "annotations",
    "depends",
    "description",
}
INITIAL_HEADERS = ["id", "urgency", "created", "modified", "tree"]

COLOR_BY_STATUS = {
    "completed": "b s black on green",
    "deleted": "s red",
    "pending": "white",
    "started": "b green",
    "recurring": "i magenta",
}

JSONDict = Dict[str, Any]


def keep_keys(keys: Iterable[str], item: JSONDict) -> JSONDict:
    """Keep only the keys in `keys` from `item`."""
    return dict(zip(keys, map(item.get, keys)))


class Annotation(TypedDict):
    created: str
    description: str


@dataclass
class Task:
    annotations: list[Annotation]
    created: str
    description: str
    id: int
    modified: str
    status: Literal["completed", "deleted", "pending", "started", "recurring"]
    urgency: int
    uuid: str

    depends: list[str] = field(default_factory=list)
    due: str | None = None
    end: str | None = None
    priority: Literal["H", "M", "L"] | None = None
    project: str | None = None
    reviewed: str | None = None
    sched: str | None = None
    start: str | None = None
    tags: list[str] = field(default_factory=list)
    wait: str | None = None

    @property
    def desc(self) -> str:
        desc = self.description
        if self.start:
            self.status = "started"

        desc = wrap(desc, COLOR_BY_STATUS[self.status])
        if self.priority:
            desc = f"{get_val(self, 'priority')} {desc}"

        return desc

    def get_tree(self, get_desc: Callable[[str], str]) -> Tree:
        tree = new_tree(title=self.desc, guide_style="white")
        if self.annotations:
            tree.add(get_val(self, "annotations"))

        dep_uuids = self.depends
        if deps := list(filter(None, map(get_desc, dep_uuids))):
            tree.add(new_tree(deps, guide_style="b red", hide_root=True))

        return tree

    def get_row(
        self, extract_data: Callable[[JSONDict], JSONDict], *args, **kwargs
    ) -> JSONDict:
        data = extract_data(asdict(self))
        data["tree"] = self.get_tree(**kwargs)

        return data


def get_headers(task_headers: Iterable[str]) -> List[str]:
    """Return the list of headers that will be used in the table."""
    ordered_keys = dict.fromkeys([*INITIAL_HEADERS, *sorted(task_headers)]).keys()
    return [*[k for k in ordered_keys if k not in SKIP_HEADERS], "tree"]


fields_map: JSONDict = {
    "id": str,
    "uuid": str,
    "urgency": lambda x: str(round(float(x), 1)),
    "description": lambda x: x,
    "due": human_dt,
    "end": human_dt,
    "sched": human_dt,
    "tags": format_with_color,
    "project": format_with_color,
    "modified": human_dt,
    "created": human_dt,
    "start": human_dt,
    "priority": lambda x: "[b]([red]![/])[/]" if x == "H" else "",
    "annotations": lambda ann: new_tree(
        (f"[b]{human_dt(a['created'])}[/]: [i]{a['description']}[/]" for a in ann),
        "Annotations",
    )
    if ann
    else None,
}
FIELDS_MAP.update(fields_map)


def get_table(tasks_data_by_group: Dict[str, list[JSONDict]], **__) -> Iterator[Panel]:
    """Yield a table for each tasks group."""
    headers = get_headers(next(iter(join(tasks_data_by_group.values()))))
    keep_headers = partial(keep_keys, headers)

    tasks_by_group = {
        g: [Task(**t) for t in tasks_data]
        for g, tasks_data in tasks_data_by_group.items()
    }
    desc_by_uuid = {t.uuid: t.desc for t in join(tasks_by_group.values())}
    for group, tasks in tasks_by_group.items():
        yield border_panel(
            flexitable([
                t.get_row(keep_headers, get_desc=desc_by_uuid.get) for t in tasks
            ]),
            title=wrap(group, "b"),
            style=predictably_random_color(group),
        )



from __future__ import annotations

import re
from collections import defaultdict
from datetime import datetime
from functools import singledispatch
from itertools import islice
from typing import TYPE_CHECKING, Any, Callable, Iterable, MutableMapping

from rich.text import Text

from .utils import (
    BOLD_GREEN,
    BOLD_RED,
    JSONDict,
    border_panel,
    duration2human,
    fmt_time,
    format_string,
    format_with_color,
    format_with_color_on_black,
    get_country,
    human_dt,
    md_panel,
    new_table,
    pretty_diff,
    progress_bar,
    simple_panel,
    split_with_color,
    sql_syntax,
    syntax,
    timestamp2timestr,
    wrap,
)

if TYPE_CHECKING:
    from rich.console import RenderableType
    from rich.table import Table


MATCH_COUNT_HEADER = re.compile(r"duration|(_sum$|_?count$)")


def counts_table(data: list[JSONDict]) -> Table:
    count_header = ""
    subcount_header = None
    ordered_headers = []
    for key in data[0]:
        if key.endswith("_subcount"):
            subcount_header = key
        elif MATCH_COUNT_HEADER.search(key):
            if count_header:
                ordered_headers.append(key)
            else:
                count_header = key
        else:
            ordered_headers.append(key)

    all_counts = [float(i[count_header]) for i in data]
    num_type = int if len({c % 1 for c in all_counts}) == 1 else float
    max_value = max(all_counts)

    if subcount_header:
        count_header = f"{subcount_header}/{count_header}".replace(
            "_count", ""
        ).replace("_subcount", "")

    table = new_table(*ordered_headers, count_header, count_header, expand=True)
    for item, count in zip(data, all_counts):
        subcount = None
        inverse = False
        count_val = str(count)
        if subcount_header:
            subcount = float(item[subcount_header])
            count_val = f"{num_type(subcount)}/{num_type(count)}"
        elif "duration" in count_header:
            inverse = True
            if num_type is int:
                count_val = duration2human(count)
        else:
            count_val = str(num_type(count))

        table.add_row(
            *(get_val(item, h) for h in ordered_headers),
            count_val,
            progress_bar(end=subcount, width=max_value, size=count, inverse=inverse),
        )
    if count_header in {"duration", "total_duration"}:
        table.caption = "Total " + duration2human(float(sum(all_counts)))
        table.caption_justify = "left"
    return table


FIELDS_MAP: MutableMapping[str, Callable[..., RenderableType]] = defaultdict(
    lambda: str,
    diff=lambda x: pretty_diff(*x),
    albumtypes=lambda x: " ".join(
        map(
            format_with_color,
            (
                ("; ".join(x) if isinstance(x, list) else x)
                .replace("compilation", "comp")
                .replace("dj-mix; broadcast", "dj-mix")
                .replace("broadcast; dj-mix", "dj-mix")
            ).split("; "),
        )
    ),
    author=format_with_color_on_black,
    labels=lambda x: (
        wrap("    ".join(wrap(y["name"].upper(), f"#{y['color']}") for y in x), "b")
        if isinstance(x, list)
        else format_with_color(x.upper())
    ),
    avg_last_played=lambda x: human_dt(x, acc=2),
    since=lambda x: (
        x
        if isinstance(x, str)
        else datetime.fromtimestamp(float(x)).strftime("%F %H:%M")
    ),
    dt=lambda x: human_dt(x, acc=5),
    wait_per_play=lambda x: wrap(
        " ".join(islice(fmt_time(int(float(x))), 1)), BOLD_GREEN
    ),
    bpm=lambda x: (
        wrap(
            str(x),
            (
                BOLD_GREEN
                if x < 135
                else BOLD_RED
                if x > 230
                else "red"
                if x > 165
                else "yellow"
            ),
        )
        if isinstance(x, int)
        else x
    ),
    length=timestamp2timestr,
    tracktotal=lambda x: (
        (wrap("{}", "b cyan") + "/" + wrap("{}", "b cyan")).format(*x)
        if isinstance(x, Iterable) and not isinstance(x, str)
        else str(x)
    ),
    category=lambda x: "/".join(map(format_with_color, x.split("/"))),
    country=get_country,
    helicopta=lambda x: ":fire: " if x and int(x) else "",
    hidden=lambda x: ":shit: " if x and int(x) else "",
    keywords=format_with_color_on_black,
    ingr=lambda x: simple_panel(format_with_color(x)),
    comments=lambda x: md_panel(
        x.replace("\n0", "\n* 0").replace("\n[", "\n* ["), title="comments"
    ),
    released=lambda x: x.replace("-00", "") if isinstance(x, str) else str(x),
    duration=lambda x: duration2human(x) if isinstance(x, (int, float)) else x,
    plays=lambda x: wrap(x, BOLD_GREEN),
    skips=lambda x: wrap(x, BOLD_RED),
    new=lambda x: (
        wrap(":heavy_check_mark: ", BOLD_GREEN)
        if x
        else wrap(":cross_mark_button: ", BOLD_RED)
    ),
    link=lambda name: (
        wrap(f" {name} ", "b black on red")
        if name == "blocks"
        else wrap(name, BOLD_RED)
        if name == "is blocked by"
        else name
    ),
    code=lambda x: syntax(x, "python"),
    context=lambda x: syntax(x, "python"),
    python=lambda x: syntax(x, "python"),
    CreatedBy=lambda x: syntax(x.replace(";", "\n"), "sh"),
    file=lambda x: "/".join(map(format_with_color, x.split("/"))),
    field=lambda x: ".".join(map(format_with_color, x.split("."))),
    unified_diff=lambda x: syntax(x, "diff"),
    diffHunk=lambda x: syntax(x, "diff"),
    snippet=lambda x: border_panel(syntax(x, "python", indent_guides=True)),
    query=lambda x: Text(x, style="bold"),
    sql=lambda x: border_panel(sql_syntax(x.replace(r"\[", "["))),
)
fields_by_func: dict[Callable[..., RenderableType], Iterable[str]] = {
    format_with_color: (
        "__typename",
        "album",
        "albumtype",
        "app",
        "area",
        "assignee",
        "brand",
        "calendar",
        "catalognum",
        "categories",
        "Category",
        "data_source",
        "default_start_time",
        "default_end_time",
        "Description",
        "endpoint",
        "entity",
        "environment",
        "epic_key",
        "epic_priority",
        "epic_status",
        "event",
        "from",
        "full_name",
        "group_source",
        "Interests",
        "issuetype",
        "key",
        "kind",
        "label",
        "mastering",
        "media",
        "module",
        "operation",
        "primary",
        "priority",
        "project",
        "short_name",
        "source",
        "status",
        "style",
        "subtask_assignee",
        "subtask_key",
        "subtask_priority",
        "subtask_status",
        "symbol",
        "table",
        "to",
        "tables",
        "test",
        "type_name",
        "user",
    ),
    split_with_color: ("genre", "genres"),
    human_dt: (
        "added",
        "committedDate",
        "due",
        "start",
        "end",
        "entry",
        "first_active",
        "last_active",
        "last_played",
        "mtime",
        "updated",
        "providerPublishTime",
    ),
    md_panel: (
        "answer",
        "benefits",
        "bodyHTML",
        "creditText",
        "description",
        "desc",
        # "instructions",
        "interview",
        "notes",
        "text",
    ),
}
for func, fields in fields_by_func.items():
    for field in fields:
        FIELDS_MAP[field] = func

DISPLAY_HEADER: dict[str, str] = {
    "track": "#",
    "bpm": "🚀",
    "last_played": ":timer_clock: ",
    "mtime": "updated",
    "data_source": "source",
    "helicopta": ":helicopter: ",
    "hidden": ":no_entry: ",
    "track_alt": ":cd: ",
    "catalognum": ":pen: ",
    "plays": wrap(":play_button: ", BOLD_GREEN),
    "skips": wrap(":stop_button: ", BOLD_RED),
    "albumtypes": "types",
}


def _get_val(value: Any, field: str) -> Any:
    if value is None:
        return "None"

    if isinstance(value, str):
        value = format_string(value)

    if isinstance(value, (int, float)):
        value = str(value)

    if field not in FIELDS_MAP and field.endswith("_group") and isinstance(value, list):
        return format_with_color(value)

    return FIELDS_MAP[field](value)


@singledispatch
def get_val(obj: JSONDict | object, field: str) -> Any:
    """Definition of a generic get_val function."""


@get_val.register
def _(obj: dict, field: str) -> Any:
    return _get_val(obj.get(field), field)


@get_val.register
def _(obj: object, field: str) -> Any:
    return _get_val(getattr(obj, field, None), field)



from __future__ import annotations

import argparse
import json
import sys
import tempfile
from functools import singledispatch
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterator, List

from rich.traceback import install
from typing_extensions import TypedDict

from . import calendar, task
from .fields import get_val
from .generic import flexitable
from .github import pulls_table
from .music import albums_table
from .utils import make_console, new_table, pretty_diff, wrap

if TYPE_CHECKING:
    from rich.console import RenderableType
    from rich.table import Table

JSONDict = Dict[str, Any]


console = make_console()
install(console=console, show_locals=True, width=console.width)


def lights_table(lights: List[JSONDict], **__) -> Iterator[Table]:
    from rgbxy import Converter

    headers = lights[0].keys()
    table = new_table(*headers)
    conv = Converter()
    for light in lights:
        xy = light.get("xy")
        style = ""
        if not light["on"]:
            style = "dim"
            light["xy"] = ""
        elif xy:
            color = conv.xy_to_hex(*xy)
            light["xy"] = wrap("   a", f"#{color} on #{color}")
        table.add_row(*(get_val(light, h) for h in headers), style=style)
    yield table


def load_data(filepath: str) -> list[JSONDict] | JSONDict | str:
    """Load data from the given file.

    Try to load the data as JSON, otherwise return the text as is.
    """
    if filepath == "/dev/stdin":
        text = sys.stdin.read()
    else:
        path = Path(filepath)
        text = path.read_text() if path.exists() and path.is_file() else filepath

    try:
        json_data: JSONDict | list[JSONDict] = json.loads(text)
    except json.JSONDecodeError:
        return text
    else:
        return json_data


def get_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="""Pretty-print JSON data.
By default, read JSON data from stdin and prettify it.
Otherwise, use command 'diff' to compare two JSON objects or blocks of text.""",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument("-j", "--json", action="store_true", help="output as JSON")
    parser.add_argument(
        "-s", "--save", action="store_true", help="save the output as HTML"
    )

    subparsers = parser.add_subparsers(
        dest="command", title="Subcommands", required=False
    )
    diff_parser = subparsers.add_parser("diff", help="show diff between two objects")
    diff_parser.add_argument(
        "before", type=load_data, help="JSON object or text before, can be a filename"
    )
    diff_parser.add_argument(
        "after", type=load_data, help="JSON object or text after, can be a filename"
    )
    return parser.parse_args()


class NamedData(TypedDict):
    title: str
    values: list[JSONDict]


TABLE_BY_NAME: Dict[str, Callable[..., Any]] = {
    "Pull Requests": pulls_table,
    "Hue lights": lights_table,
    "Calendar": calendar.get_table,
    "Album": albums_table,
    "Tasks": task.get_table,
}


@singledispatch
def draw_data(data: Any, **kwargs) -> Iterator[RenderableType]:
    """Render the provided data."""
    yield data


@draw_data.register(dict)
def _draw_data_dict(data: JSONDict | NamedData, **kwargs) -> Iterator[RenderableType]:
    if (title := data.get("title")) and (values := data.get("values")):
        table = TABLE_BY_NAME.get(title, flexitable)
        yield from table(values, **kwargs)
    else:
        yield flexitable(data)


@draw_data.register(list)
def _draw_data_list(data: list[JSONDict], **kwargs) -> Iterator[RenderableType]:
    yield flexitable(data)


def main() -> None:
    args = get_args()
    if args.command == "diff":
        console.print(pretty_diff(args.before, args.after))
    else:
        data = load_data("/dev/stdin")
        if args.json:
            console.print_json(data=data)
        else:
            console.record = True
            for renderable in filter(None, draw_data(data, verbose=args.verbose)):
                console.print(renderable)

    if args.save:
        filename = tempfile.NamedTemporaryFile(suffix=".html", delete=False).name
        console.save_html(filename)
        print(f"Saved output as {filename}", file=sys.stderr)


if __name__ == "__main__":
    main()



from __future__ import annotations

import itertools as it
import logging
import os
from datetime import datetime
from functools import partial, wraps
from typing import Any, Callable, Dict, Iterable, List, TypeVar, Union

from multimethod import multidispatch
from rich import box
from rich.columns import Columns
from rich.console import ConsoleRenderable, RenderableType
from rich.logging import RichHandler
from rich.markdown import Markdown
from rich.panel import Panel
from rich.text import Text
from rich.tree import Tree

from . import fields
from .fields import DISPLAY_HEADER, MATCH_COUNT_HEADER, _get_val, counts_table
from .utils import (
    NewTable,
    border_panel,
    format_with_color,
    group_by,
    list_table,
    make_console,
    new_table,
    new_tree,
    predictably_random_color,
    wrap,
)

JSONDict = Dict[str, Any]
T = TypeVar("T")
console = make_console()

indent = ""

MAX_DICT_LENGTH = int(os.getenv("TABLE_MAX_DICT_LENGTH") or 500)


def time_fmt(current: datetime) -> Text:
    diff = current - time_fmt.prev  # type: ignore [attr-defined]
    time_fmt.prev = current  # type: ignore [attr-defined]
    return Text(f"{diff.total_seconds() * 100:.2f}s")


time_fmt.prev = datetime.now()  # type: ignore [attr-defined]


log = logging.getLogger(__name__)
if not log.handlers:
    handler = RichHandler(
        tracebacks_show_locals=True,
        omit_repeated_times=False,
        show_time=True,
        log_time_format=time_fmt,
    )
    log.addHandler(handler)
    if os.getenv("DEBUG"):
        log.setLevel("DEBUG")


def _debug(_func: Callable[..., T], *args) -> None:
    if log.isEnabledFor(10):
        global indent
        data, *header = (str(arg).split(r"\n")[0] for arg in args)
        print(
            indent
            + " ".join(
                [
                    f"Function \033[1;31m{_func.__name__}\033[0m",
                    f"Types: \033[1;33m{list(_func.__annotations__.values())[:-1]}\033[0m",
                    f"Header: \033[1;35m{header[0]}\033[0m " if header else "",
                    f"Data: \033[1m{data}\033[0m" if data else "",
                ]
            )
        )

        indent += "│ "


def _undebug(_type: type, *args: Any) -> None:
    global indent
    indent = indent[:-2]
    if log.isEnabledFor(10):
        print(indent + "└─ " + f"Returning {_type}")


def mapping_view_table() -> NewTable:
    """A table with two columns
    * First for bold field names
    * Second one for values.
    """
    table = new_table(border_style="cyan", style="cyan", box=box.MINIMAL, expand=False)
    table.add_column(justify="right", style="bold misty_rose1")
    table.add_column()
    return table


def prepare_dict(item: JSONDict) -> JSONDict:
    if "before" in item and "after" in item:
        item["diff"] = (item.pop("before"), item.pop("after"))
    return item


def debug(func: Callable[..., T]) -> Callable[..., T]:
    @wraps(func)
    def wrapper(*args: Any) -> T:
        _debug(func, *args)
        result = func(*args)
        _undebug(type(result), *args)
        return result

    return wrapper


@multidispatch
@debug
def flexitable(data) -> RenderableType:
    return str(data)


@flexitable.register
@debug
def _header(data: Any, header: str) -> RenderableType:
    if data in ("", [], {}):
        return ""

    if header not in fields.FIELDS_MAP or isinstance(data, (dict, list)):
        return flexitable(data)

    out = _get_val(data, header)
    if not isinstance(out, str) and type(out) == type(data):
        return flexitable(out)

    return out


@flexitable.register
@debug
def _tuple_header(data: tuple, header: str) -> RenderableType:  # type: ignore[type-arg]
    return fields.FIELDS_MAP[header](data) if header in fields.FIELDS_MAP else str(data)


@flexitable.register
@debug
def _renderable(data: Union[ConsoleRenderable, NewTable]) -> RenderableType:
    return data


@flexitable.register
@debug
def _str(data: str) -> RenderableType:
    return data


@flexitable.register
@debug
def _json_dict(data: dict) -> RenderableType:
    data = prepare_dict(data)
    table = mapping_view_table()
    cols: List[RenderableType] = []
    for key, content in data.items():
        if content is None or (isinstance(content, list) and not content):
            continue

        value = flexitable(content, key)
        # print(f"{key=}, {value=}")
        if isinstance(value, (NewTable, Text, dict, Columns)):
            cols.append(border_panel(value, title=key))
        elif isinstance(value, ConsoleRenderable) and hasattr(value, "title"):
            value.title = key
            cols.append(value)
        elif isinstance(value, ConsoleRenderable) and not isinstance(value, Markdown):
            cols.append(value)
        else:
            table.add_row(key, value)

    if table.rows:
        cols.insert(0, table)

    if len(cols) == 1:
        return cols[0]

    row: List[RenderableType] = []
    width = 0
    rows: List[RenderableType] = []
    for rend in cols:
        this_width = console.measure(rend).maximum
        if width + this_width > console.width:
            rows.append(Columns(row, equal=True, padding=(0, 0)))
            row, width = [rend], this_width
        else:
            row.append(rend)
            width += this_width

    rows.append(Columns(row, equal=True, padding=(0, 0)))

    return list_table(rows, padding=(0, 0))


simple_head_table = partial(
    new_table, expand=False, box=box.SIMPLE_HEAD, border_style="cyan"
)


@flexitable.register
@debug
def _list(data: List) -> RenderableType:
    return flexitable(tuple(data))


@flexitable.register
@debug
def _str_list(data: Iterable[Any]) -> RenderableType:
    return format_with_color(data)


@flexitable.register
@debug
def _int_list(data: Iterable[int]) -> Columns:
    return Columns(str(x) for x in data)


@flexitable.register
@debug
def _dict_list(data: Iterable[dict]) -> RenderableType:
    if len(data) == 1 and len(data[0]) > 10:
        return flexitable(data[0])

    data = [prepare_dict(item) for item in data if item]
    all_keys = dict.fromkeys(it.chain.from_iterable(tuple(d.keys()) for d in data))
    if not all_keys:
        return simple_head_table([])

    keys = {
        k: None for k in all_keys if any((d.get(k) is not None) for d in data)
    }.keys()

    overlap = set(map(type, data[0].values())) & {int, float, str}

    if overlap and any(MATCH_COUNT_HEADER.search(k) for k in keys):
        return counts_table(data)

    def getval(value: Any, key: str) -> RenderableType:
        transformed_value = flexitable(value, key)
        header = wrap(key, "b")
        if (
            isinstance(transformed_value, NewTable)
            and len(transformed_value.rows) == 1
            and len(transformed_value.columns) == 1
        ):
            transformed_value = transformed_value.columns[0]._cells[0]

        if isinstance(transformed_value, str):
            return f"{header}: {transformed_value}"

        if isinstance(transformed_value, Panel):
            transformed_value.title = header
        elif isinstance(transformed_value, Tree):
            transformed_value.label = header

        return transformed_value

    large_table = simple_head_table()
    for large, items in group_by(
        data, lambda i: len(str(i.values())) > MAX_DICT_LENGTH
    ):
        if large:
            for item in items:
                values = it.starmap(
                    getval,
                    sorted(
                        [(v or "", k) for k, v in item.items() if k in keys],
                        key=lambda x, *_: str(type(x)),
                        reverse=True,
                    ),
                )
                tree = new_tree(values, "")
                large_table.add_row(tree)
        else:
            sub_table = simple_head_table(show_header=True)
            for key in keys:
                sub_table.add_column(key, header_style=predictably_random_color(key))
            for item in items:
                sub_table.add_dict_item(item, transform=flexitable)
            for col in sub_table.columns:
                col.header = DISPLAY_HEADER.get(str(col.header), col.header)
            large_table.add_row(sub_table)

    return large_table



from __future__ import annotations

import operator as op
from collections import defaultdict
from functools import lru_cache, partial
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Tuple, TypeVar

from rich import box
from rich.align import Align
from rich.console import ConsoleRenderable, Group

from .fields import DISPLAY_HEADER, FIELDS_MAP
from .utils import (
    NewTable,
    border_panel,
    group_by,
    new_table,
    predictably_random_color,
    simple_panel,
    wrap,
)

if TYPE_CHECKING:
    from rich.panel import Panel
    from rich.table import Table

JSONDict = Dict[str, Any]

TRACK_FIELDS = [
    "track",
    "length",
    "artist",
    "title",
    "bpm",
    "last_played",
    "plays",
    "skips",
    "helicopta",
    "hidden",
]
ALBUM_IGNORE = set(TRACK_FIELDS) | {
    "album_color",
    "albumartist_color",
    "album",
    "album_title",
    "comments",
    "genre",
    "tracktotal",
    "albumartist",
}


new_table = partial(new_table, collapse_padding=True, expand=True, padding=0)


def get_header(key: str) -> str:
    return DISPLAY_HEADER.get(key, key)


@lru_cache(maxsize=128)
def get_val(track: JSONDict, field: str) -> Any:
    trackdict = dict(track)
    return FIELDS_MAP[field](trackdict[field]) if trackdict.get(field) else ""


def get_vals(
    fields: Iterable[str], tracks: Iterable[JSONDict]
) -> Iterable[Iterable[str]]:
    return [[get_val(tuple(t.items()), f) for f in fields] for t in tracks]


def tracks_table(tracks: List[JSONDict], fields: List[str], color: str) -> NewTable:
    return new_table(
        *map(get_header, fields),
        rows=get_vals(fields, tracks),
        border_style=color,
        padding=(0, 0, 0, 1),
    )


T = TypeVar("T")


def album_stats(tracks: List[JSONDict]) -> JSONDict:
    def agg(field: str, default: T) -> Iterable[T]:
        return ((x.get(field) or default) for x in tracks)

    stats: JSONDict = dict(
        bpm=round(sum(agg("bpm", 0)) / len(tracks)),
        rating=round(sum(agg("rating", 0)) / len(tracks), 2),
        plays=sum(agg("plays", 0)),
        skips=sum(agg("skips", 0)),
        mtime=max(agg("mtime", 0)),
        last_played=max(agg("last_played", 0)),
        tracktotal=(str(len(tracks)), str(tracks[0].get("tracktotal")) or "0"),
        comments="\n---\n---\n".join(set(agg("comments", ""))),
    )
    return stats


def add_colors(album: JSONDict) -> None:
    for field in "album", "albumartist":
        val = (album.get(field) or "").replace("Various Artists", "VA")
        color = predictably_random_color(val)
        album[f"{field}_color"] = color
        val = album.get(field)
        album[field] = wrap(val, f"b i {color}") if val else ""


def format_title(title: str) -> str:
    return wrap(f"  {title}  ", "i white on grey3")


def album_title(album: JSONDict) -> Table:
    name = album["album"]
    artist = album.get("albumartist") or album.get("artist")
    genre = album.get("genre") or ""
    released = album.get("released", "")
    title = (
        f"{name} by {artist}"
        if album["albumtypes"] != "single"
        else f"singles by {artist}"
    )
    return new_table(rows=[[format_title(title), format_title(released), genre]])


def album_info(tracks: List[JSONDict]) -> JSONDict:
    first = tracks[0]
    fields = sorted([f for f in tracks[0] if f not in TRACK_FIELDS])

    album = defaultdict(str, zip(fields, op.itemgetter(*fields)(first)))
    if not album["album"]:
        album.update(album="Singles", albumartist=first["artist"])
    album.update(**album_stats(tracks))
    add_colors(album)
    for field, _ in filter(op.truth, sorted(album.items())):
        album[field] = get_val(tuple(album.items()), field)
    album["album_title"] = album_title(album)
    return album


def album_info_table(album: JSONDict) -> Table:
    def should_display(keyval: Tuple[str, Any]) -> bool:
        return keyval[1] and keyval[0] not in ALBUM_IGNORE

    items = filter(should_display, sorted(album.items()))
    table = new_table(rows=((get_header(x[0]), x[1]) for x in items))
    table.columns[0].style = "b " + album["album_color"]
    return table


def album_panel(tracks: List[JSONDict]) -> Panel:
    album = album_info(tracks)
    url = album.pop("url")

    # ignore the artist field if there is only one found
    track_fields = [
        t
        for t in TRACK_FIELDS
        if len(tracks) > 1 or len({t.get("artist") for t in tracks}) == 1
    ]
    tracks = sorted(tracks, key=op.itemgetter("track", "artist", "title"))
    tracklist = tracks_table(tracks, track_fields, album["album_color"])

    last_track_index = tracks.index(
        next(iter(sorted(tracks, key=op.itemgetter("last_played"), reverse=True)))
    )
    tracklist.rows[last_track_index].style = "b white on #000000"
    tracklist.add_row(
        *[album.get(k) or "" for k in ["tracktotal", *track_fields[1:]]],
        style="d white on grey11",
    )

    comments = album.get("comments")
    return border_panel(
        Group(
            album["album_title"],
            Align.center(
                simple_panel(comments, style="grey54", expand=True, align="center")
            )
            if comments
            else "",
            new_table(rows=[[album_info_table(album), tracklist]]),
        ),
        box=box.DOUBLE_EDGE,
        style=album["albumartist_color"],
        subtitle=f"[b white]{url}",
    )


def albums_table(all_tracks: List[JSONDict], **__) -> Iterable[ConsoleRenderable]:
    def get_album(track: JSONDict) -> str:
        return track.get("album") or ""

    for track in all_tracks:
        if not track["album"] and "single" in track.get("albumtype", ""):
            track["album"] = "singles"
            track["albumartist"] = track["label"]

    for _, tracks in group_by(all_tracks, get_album):
        yield album_panel(list(tracks))



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timedelta
from operator import attrgetter
from typing import TYPE_CHECKING, Any, Dict, Iterable

from funcy import join
from rich.bar import Bar
from rich.columns import Columns
from typing_extensions import Literal, NotRequired, TypedDict

from .fields import get_val
from .utils import (
    border_panel,
    group_by,
    new_table,
    wrap,
)

if TYPE_CHECKING:
    from rich.console import RenderableType

JSONDict = Dict[str, Any]

PAST_COLOR = "grey7"
SYMBOL_BY_STATUS = {
    "needsAction": "[b grey3] ? [/]",
    "accepted": "[b green] ✔ [/]",
    "declined": "[b red] ✖ [/]",
    "tentative": "[b yellow] ? [/]",
}


class Datetime(TypedDict):
    dateTime: NotRequired[str]
    date: NotRequired[str]
    timeZone: NotRequired[str]


@dataclass
class Period:
    color: str
    desc: str
    end: datetime
    summary: str
    start: datetime
    status_symbol: str

    @classmethod
    def make(cls, **kwargs) -> Period:
        if kwargs["end"].replace(tzinfo=None) < datetime.now():
            kwargs["color"] = PAST_COLOR
        return cls(**kwargs)

    @property
    def start_day(self) -> str:
        return self.start.strftime("%d %a")

    @property
    def start_year_month(self) -> str:
        return self.start.strftime("%Y %B")

    @property
    def start_time(self) -> str:
        return wrap(self.start.strftime("%H:%M"), "white")

    @property
    def end_time(self) -> str:
        return wrap(self.end.strftime("%H:%M"), "white")

    @property
    def bar(self) -> Bar:
        if self.start.hour == self.end.hour == 0:
            begin, end = 0, 86400
        else:
            midnight_ts = self.start.replace(hour=0).timestamp()
            begin = int(self.start.timestamp() - midnight_ts)
            end = int(self.end.timestamp() - midnight_ts)

        return Bar(86400, begin, end, color=self.color)

    @property
    def name(self) -> RenderableType:
        title = self.status_symbol + wrap(self.summary, f"b {self.color}")
        return border_panel(get_val(self, "desc"), title=title) if self.desc else title


@dataclass
class Event:
    backgroundColor: str
    calendar: str
    desc: str
    end: datetime
    summary: str
    status: Literal["accepted", "needsAction", "declined", "tentative"]
    start: datetime

    @classmethod
    def make(cls, start: Datetime, end: Datetime, summary: str, **kwargs) -> Event:
        return cls(
            start=cls.get_datetime(start),
            end=cls.get_datetime(end),
            summary=summary or "busy",
            **kwargs,
        )

    @property
    def status_symbol(self) -> str:
        return SYMBOL_BY_STATUS[self.status]

    @staticmethod
    def get_datetime(date_obj: Datetime) -> datetime:
        date = date_obj.get("dateTime") or date_obj.get("date") or ""
        return datetime.fromisoformat(date.strip("Z"))

    def get_periods(self) -> list[Period]:
        diff = self.end - self.start
        h_after_midnight = (24 * diff.days + (diff.seconds // 3600)) - (
            24 - self.start.hour
        )

        end = (self.start + timedelta(days=1)).replace(hour=0, minute=0, second=0)

        def eod(day_offset: int) -> datetime:
            return (self.start + timedelta(days=day_offset)).replace(
                hour=23, minute=59, second=59
            )

        def midnight(day_offset: int) -> datetime:
            return (self.start + timedelta(days=day_offset)).replace(
                hour=0, minute=0, second=0
            )

        days_count = h_after_midnight // 24 + 1
        periods = []
        for start, end in zip(
            [self.start, *map(midnight, range(1, days_count + 1))],
            [*map(eod, range(days_count)), self.end],
        ):
            periods.append(
                Period.make(
                    status_symbol=self.status_symbol,
                    color=self.backgroundColor,
                    start=start,
                    end=end,
                    desc=self.desc,
                    summary=self.summary,
                )
            )
        return periods


def get_legend(events: list[Event]) -> RenderableType:
    """Return a renderable with a list of calendars and their colors."""
    calendar_and_color = sorted({(e.calendar, e.backgroundColor) for e in events})
    colored_calendars = (
        wrap(f" {c} ", f"b black on {clr}") for c, clr in calendar_and_color
    )
    return Columns(
        colored_calendars,
        title="Calendars",
        expand=True,
        equal=True,
        align="center",
    )


def get_months(events: list[Event]) -> Iterable[RenderableType]:
    all_periods = join(e.get_periods() for e in events)

    headers = "name", "start_time", "end_time", "bar"
    get_values = attrgetter(*headers)
    for year_and_month, month_periods in group_by(
        all_periods, lambda x: x.start_year_month
    ):
        table = new_table(*headers, highlight=False, padding=0, show_header=False)
        for day, day_periods in group_by(month_periods, lambda x: x.start_day):
            table.add_row(wrap(day, "b i"))
            for period in day_periods:
                values = get_values(period)
                if "Week " in period.summary:
                    table.add_row("")
                    table.add_row(*values, style=period.color + " on grey7")
                else:
                    table.add_row(*values)
            table.add_row("")
        yield border_panel(table, title=year_and_month)


def get_table(events_data: list[JSONDict], **__) -> Iterable[RenderableType]:
    events = [Event.make(**e) for e in events_data]
    yield get_legend(events)
    yield from get_months(events)



"""Functionality to display data from GitHub API."""

from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from itertools import chain
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    List,
    Mapping,
    Protocol,
    Union,
)

from rich import box
from rich.syntax import Syntax
from typing_extensions import TypedDict

from .fields import FIELDS_MAP, _get_val, get_val
from .generic import flexitable
from .utils import (
    JSONDict,
    border_panel,
    diff_dt,
    format_with_color,
    format_with_color_on_black,
    group_by,
    list_table,
    md_panel,
    new_table,
    predictably_random_color,
    simple_panel,
    wrap,
)

if TYPE_CHECKING:
    from rich.console import ConsoleRenderable, RenderableType
    from rich.panel import Panel
    from rich.table import Table


def b_green(text: str) -> str:
    """Make the text bold green."""
    return wrap(text, "b green")


def b_red(text: str) -> str:
    """Make the text bold red."""
    return wrap(text, "b red")


def fmt_add_del(added: int, deleted: int) -> List[str]:
    """Format added and deleted diff counts."""
    additions = f"+{added}" if added else ""
    deletions = f"-{deleted}" if deleted else ""
    return [b_green(additions.rjust(5)), b_red(deletions.rjust(3))]


COLOR_BY_STATE = defaultdict(
    lambda: "default",
    {
        "True": "green",
        True: "green",
        "APPROVED": "green",
        "RESOLVED": "s green",
        "OPEN": "green",
        "MERGED": "magenta",
        "CLOSED": "b red",
        "PENDING": "yellow",
        "OUTDATED": "yellow",
        "COMMENTED": "yellow",
        "CHANGES_REQUESTED": "b red",
        "REVIEW_REQUIRED": "red",
        "DISMISSED": "gray42",
        "False": "red",
    },
)


def fmt_state(state: str) -> str:
    return wrap(state, f"b {COLOR_BY_STATE[state]}")


def diff_panel(title: str, rows: List[List[str]]) -> Panel:
    return border_panel(
        new_table(rows=rows),
        title=title,
        border_style=f"dim {predictably_random_color(title)}",
    )


PR_FIELDS_MAP: Mapping[str, Callable[..., RenderableType]] = {
    "statusCheckRollup": lambda x: {
        "SUCCESS": ":green_square:",
        "FAILURE": ":red_square:",
        "PENDING": ":yellow_square:",
        "None": "",
    }[x],
    "state": lambda x: wrap(fmt_state(x), "b"),
    "reviewDecision": lambda x: wrap(fmt_state(x), "b"),
    "dates": lambda x: new_table(
        rows=[
            [b_green(r" ⬤ "), diff_dt(x[0])],
            [wrap(r" ◯ ", "b yellow"), diff_dt(x[1])],
        ]
    ),
    "path": lambda x: wrap(x, "b"),
    "message": lambda x: wrap(x, "i"),
    "files": lambda files: diff_panel(
        "files",
        [
            [*fmt_add_del(f["additions"], f["deletions"]), get_val(f, "path")]
            for f in files
        ],
    ),
    "reviewRequests": format_with_color_on_black,
    "participants": lambda x: "\n".join(
        map(format_with_color, map("{:^20}".format, x))
    ),
}


class Diff(TypedDict):
    additions: int
    deletions: int


class File(Diff):
    path: str


@dataclass
class Commit:
    additions: int
    deletions: int
    committedDate: str
    message: str
    statusCheckRollup: str

    @property
    def diff(self) -> List[str]:
        return fmt_add_del(self.additions, self.deletions)

    @property
    def parts(self) -> List[str]:
        return [
            *self.diff,
            get_val(self, "statusCheckRollup"),
            get_val(self, "message"),
            get_val(self, "committedDate"),
        ]


@dataclass
class Commits:
    commits: List[Commit]

    @property
    def panel(self) -> Panel:
        return diff_panel("commits", [commit.parts for commit in self.commits])


@dataclass
class Reaction:
    user: str
    content: str

    def __str__(self) -> str:
        return f":{self.content.lower()}: {_get_val(self.user, 'author')}"


class CreatedMixin(Protocol):
    @property
    def created(self) -> str:
        raise NotImplementedError


@dataclass
class PanelMixin:
    def get_title(self, fields: List[str]) -> str:
        return " ".join(get_val(self, f) for f in fields)

    @property
    def panel(self) -> Panel:
        raise NotImplementedError


@dataclass
class CreatedPanelMixin(CreatedMixin, PanelMixin):
    pass


@dataclass
class Content(CreatedPanelMixin):
    createdAt: str
    author: str
    body: str

    @property
    def created_at(self) -> str:
        return f"""[white]{self.createdAt.replace("T", " ").replace("Z", "")}[/]"""

    @property
    def created(self) -> str:
        return self.created_at


@dataclass
class Comment(Content):
    reactions: List[Reaction]

    @classmethod
    def make(cls, reactions: List[JSONDict], **kwargs: Any) -> Comment:
        kwargs["reactions"] = [Reaction(**c) for c in reactions]
        return cls(**kwargs)

    @property
    def title(self) -> str:
        return self.get_title(["author", "created_at"])

    @property
    def subtitle(self) -> str:
        return (
            " ".join(map(str, self.reactions))
            .replace(":laugh:", ":laughing:")
            .replace(":hooray:", ":party_popper:")
        )


@dataclass
class IssueComment(Comment):
    @property
    def panel(self) -> Panel:
        return border_panel(
            list_table([md_panel(self.body)]),
            border_style="b yellow",
            title=self.title,
            subtitle=self.subtitle,
            box=box.ROUNDED,
        )


@dataclass
class ReviewComment(Comment):
    outdated: bool
    path: str
    diffHunk: str
    pullRequestReview: str

    @property
    def diff(self) -> Syntax:
        return Syntax(
            self.diffHunk, "diff", theme="paraiso-dark", background_color="black"
        )

    @property
    def review_id(self) -> str:
        return self.pullRequestReview

    @property
    def panel(self) -> Panel:
        return md_panel(
            self.body.replace("suggestion", "python"),
            title=self.title,
            subtitle=self.subtitle,
        )


class ResolvedMixin:
    @property
    def resolved(self) -> bool:
        raise NotImplementedError

    @property
    def border_color(self) -> str:
        return "green" if self.resolved else "yellow"


@dataclass
class ReviewThread(CreatedPanelMixin, ResolvedMixin):
    path: str
    isResolved: bool
    isOutdated: bool
    resolvedBy: str
    comments: List[ReviewComment]

    @property
    def resolved(self) -> bool:
        return self.isResolved

    @classmethod
    def make(cls, comments: List[JSONDict], **kwargs: Any) -> ReviewThread:
        kwargs["comments"] = [ReviewComment.make(**c) for c in comments]
        return cls(**kwargs)

    @property
    def review_id(self) -> str:
        return self.comments[0].review_id

    @property
    def created(self) -> str:
        return self.comments[0].created_at

    @property
    def formatted_state(self) -> str:
        return (
            (
                fmt_state("RESOLVED")
                + wrap(" by ", "white")
                + format_with_color(self.resolvedBy)
            )
            if self.isResolved
            else fmt_state("PENDING")
        )

    @property
    def title(self) -> str:
        return " ".join([
            wrap(self.path, "b magenta"),
            self.formatted_state,
            fmt_state("OUTDATED") if self.isOutdated else "",
        ])

    @property
    def panel(self) -> Panel:
        comments = self.comments
        comments_col = list_table((c.panel for c in comments), padding=(1, 0, 0, 0))
        return border_panel(
            new_table(
                rows=[[self.comments[0].diff, simple_panel(comments_col)]],
                highlight=False,
            ),
            highlight=False,
            border_style=self.border_color,
            title=self.title,
        )


@dataclass
class Review(Content, ResolvedMixin):
    id: str
    state: str
    threads: List[ReviewThread]
    comments: List[ReviewComment]

    verbose: bool

    @property
    def resolved(self) -> bool:
        return all(t.resolved for t in self.threads)

    @property
    def panel(self) -> Panel:
        rows = []
        if self.body:
            rows.append(md_panel(self.body))

        if not self.resolved or self.verbose:
            self.threads.sort(key=lambda t: t.isResolved)
            rows.extend(t.panel for t in self.threads)

        return border_panel(
            list_table(rows),
            subtitle=self.state,
            border_style=self.border_color,
            title=self.title,
            box=box.HEAVY,
        )

    @property
    def status(self) -> str:
        resolved_count = sum(t.isResolved for t in self.threads)
        total_count = len(self.threads)
        if not total_count:
            return ""

        if total_count == resolved_count:
            return wrap("RESOLVED", COLOR_BY_STATE["RESOLVED"])

        return b_green(" ⬤ " * resolved_count) + b_red(
            " ◯ " * (total_count - resolved_count)
        )

    @property
    def title(self) -> str:
        return self.get_title(["state", "author", "created_at", "status"])


@dataclass
class PullRequest:
    id: str
    createdAt: str
    author: str
    body: str
    additions: int
    commits: Commits
    deletions: int
    files: List[File]
    headRefName: str
    labels: List[str]
    participants: List[str]
    repository: str
    reviewDecision: str
    reviewRequests: List[str]
    reviewThreads: List[ReviewThread]
    reviews: List[Review]
    comments: List[IssueComment]
    state: str
    title: str
    updatedAt: str
    url: str

    verbose: bool

    @property
    def dates(self) -> tuple[str, str]:
        return self.createdAt, self.updatedAt


@dataclass
class PullRequestTable(PullRequest):
    @classmethod
    def make(cls, reviews: List[JSONDict], **kwargs: Any) -> PullRequestTable:
        kwargs["commits"] = Commits([Commit(**c) for c in kwargs["commits"]])
        kwargs["comments"] = [IssueComment.make(**c) for c in kwargs["comments"]]
        threads = [ReviewThread.make(**rt) for rt in kwargs["reviewThreads"]]
        review_comments = list(chain.from_iterable(t.comments for t in threads))
        review_comments.sort(key=lambda c: c.review_id)
        comments_by_review_id = dict(group_by(review_comments, lambda c: c.review_id))
        threads.sort(key=lambda t: t.review_id)
        threads_by_review_id = dict(group_by(threads, lambda t: t.review_id))
        kwargs["reviews"] = [
            Review(
                **r,
                threads=threads_by_review_id.pop(r["id"], []),
                comments=comments_by_review_id.pop(r["id"], []),
                verbose=kwargs["verbose"],
            )
            for r in reviews
            if (r["id"] in threads_by_review_id or r["body"])
        ]
        return cls(**kwargs)

    def make_info_subpanel(self, attr: str) -> Panel:
        return simple_panel(
            get_val(self, attr),
            title=wrap(attr, "b"),
            title_align="center",
            expand=True,
            align="center",
        )

    @property
    def name(self) -> str:
        return wrap(self.title, f"b {predictably_random_color(self.title)}")

    @property
    def repo(self) -> str:
        return wrap(self.repository, f"b {predictably_random_color(self.repository)}")

    @property
    def pr_state(self) -> str:
        return "MERGED" if self.state == "MERGED" else self.reviewDecision

    @property
    def info(self) -> Panel:
        fields = "author", "dates", "headRefName", "participants", "reviewRequests"
        pairs = [(f, getattr(self, f)) for f in fields]
        field_rows = [[flexitable({f: v})] for f, v in pairs if v]
        return border_panel(
            new_table(rows=[*field_rows, [md_panel(self.body)], [self.files_commits]]),
            title=f"{self.name} @ {self.repo}",
            box=box.DOUBLE_EDGE,
            border_style=COLOR_BY_STATE[self.pr_state],
            subtitle=(
                wrap(
                    " ".join([
                        fmt_state(self.reviewDecision),
                        wrap("//", "white"),
                        fmt_state(self.state),
                    ]),
                    "b",
                )
            ),
            align="center",
            title_align="center",
            subtitle_align="center",
        )

    @property
    def files_commits(self) -> Table:
        return new_table(rows=[[get_val(self, "files"), self.commits.panel]])

    @property
    def timestamped_contents(self) -> List[CreatedPanelMixin]:
        return [*self.reviews, *self.comments]

    @property
    def panels(self) -> Iterable[Panel]:
        for content in sorted(self.timestamped_contents, key=lambda c: c.created):
            yield content.panel


def pulls_table(
    data: List[Mapping[str, Any]], **kwargs
) -> Iterable[Union[str, ConsoleRenderable]]:
    FIELDS_MAP.update(PR_FIELDS_MAP)

    pr = data[0]
    pr_table = PullRequestTable.make(**pr, verbose=kwargs["verbose"])
    yield pr_table.info
    yield from pr_table.panels



from __future__ import annotations

import random
import re
import time
from datetime import datetime, timedelta, timezone
from difflib import SequenceMatcher
from itertools import groupby, islice, starmap, zip_longest
from math import copysign
from pprint import pformat
from string import ascii_uppercase, punctuation
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Match,
    Optional,
    Protocol,
    Sequence,
    SupportsFloat,
    Tuple,
    TypeVar,
    Union,
)

import platformdirs
import sqlparse
from multimethod import multimethod
from rich import box
from rich.align import Align
from rich.bar import Bar
from rich.console import Console, RenderableType
from rich.markdown import Markdown
from rich.panel import Panel
from rich.syntax import Syntax
from rich.table import Table
from rich.text import Text
from rich.theme import Theme
from rich.tree import Tree

JSONDict = Dict[str, Any]
SPLIT_PAT = re.compile(r"[;,] ?")
PRED_COLOR_PAT = re.compile(r"(pred color)\]([^\[]+)")
HTML_PARAGRAPH = re.compile(r"</?p>")


BOLD_GREEN = "b green"
BOLD_RED = "b red"
SECONDS_PER_DAY = 86400
CONSECUTIVE_SPACE = re.compile("(?:^ +)|(?: +$)")


_T_contra = TypeVar("_T_contra", contravariant=True)


class SupportsDunderLT(Protocol[_T_contra]):
    def __lt__(self, __other: _T_contra) -> bool:
        pass


T = TypeVar("T")
K = TypeVar("K", bound=SupportsDunderLT[Any])


def group_by(iterable: Iterable[T], key: Callable[[T], K]) -> List[Tuple[K, List[T]]]:
    return [(k, list(g)) for k, g in groupby(sorted(iterable, key=key), key)]


def format_string(text: str) -> str:
    if "pred color]" in text:
        return PRED_COLOR_PAT.sub(fmt_pred_color, text)
    if "[" in text and r"\[" not in text and "[/" not in text:
        return text.replace("[", r"\[")

    return text


def wrap(text: str, tag: str) -> str:
    return f"[{tag}]{format_string(str(text))}[/]"


def format_space(string: str) -> str:
    return CONSECUTIVE_SPACE.sub(r"[u]\g<0>[/]", string)


def format_new(string: str) -> str:
    return wrap(format_space(string), BOLD_GREEN)


def format_old(string: str) -> str:
    return wrap(wrap(string, BOLD_RED), "s")


def fmtdiff(change: str, before: str, after: str) -> str:
    if change == "insert":
        return format_new(after)
    if change == "delete":
        return format_old(before)
    if change == "replace":
        return format_old(before) + format_new(after)

    return wrap(before, "dim")


def make_difftext(
    before: str,
    after: str,
    junk: str = "".join(
        sorted((set(punctuation) - {"_", "-", ":"}) | set(ascii_uppercase))
    ),
) -> str:
    matcher = SequenceMatcher(
        lambda x: x not in junk, autojunk=False, a=before, b=after
    )
    diff = ""
    for code, a1, a2, b1, b2 in matcher.get_opcodes():
        diff += fmtdiff(code, before[a1:a2], after[b1:b2]) or ""
    return diff


def duration2human(duration: SupportsFloat) -> str:
    diff = timedelta(seconds=float(duration))
    days = f"{diff.days}d " if diff.days else ""
    time_parts = [diff.seconds // 3600, diff.seconds % 3600 // 60, diff.seconds % 60]
    return "{:>12}".format(days + ":".join(map("{0:0>2}".format, time_parts)))


def fmt_time(seconds: int) -> Iterable[str]:
    abs_seconds = abs(seconds)
    return (
        f"{int(copysign(num, seconds)):>3}{unit}"
        for num, unit in (
            (abs_seconds // 86400, "d"),
            (abs_seconds // 3600, "h"),
            (abs_seconds % 3600 // 60, "m"),
            (abs_seconds % 60, "s"),
        )
        if num
    )


def get_theme() -> Optional[Theme]:
    config_path = platformdirs.user_config_path("rich") / "config.ini"
    if config_path.exists():
        return Theme.read(str(config_path))
    return None


def make_console(**kwargs: Any) -> Console:
    return Console(
        theme=get_theme(),
        force_terminal=True,
        force_interactive=False,
        emoji=True,
        **kwargs,
    )


console = make_console()
print = console.print


class NewTable(Table):
    def __init__(self, *args: str, **kwargs: Any) -> None:
        ckwargs = {
            "overflow": kwargs.pop("overflow", "fold"),
            "justify": kwargs.pop("justify", "left"),
            "vertical": kwargs.pop("vertical", "middle"),
        }
        super().__init__(**kwargs)
        for arg in args:
            self.add_column(arg, **ckwargs)

    def add_rows(self, rows: Iterable[Iterable[RenderableType]]) -> None:
        """Add multiple rows to the table."""
        for row in rows:
            self.add_row(*row)

    @property
    def colnames(self) -> List[str]:
        """Provide a mapping between columns names / ids and columns."""
        return [str(c.header) for c in self.columns]

    def add_dict_item(
        self,
        item: JSONDict,
        transform: Callable[[Any, str], Any] = lambda x, _: x,
        **kwargs: Any,
    ) -> None:
        """Take the required columns / keys from the given dictionary item."""
        vals = (transform(item.get(c, ""), c) for c in self.colnames)
        self.add_row(*vals, **kwargs)


def new_table(*headers: str, **kwargs: Any) -> NewTable:
    default = {
        "show_edge": False,
        "show_header": False,
        "pad_edge": False,
        "highlight": True,
        "row_styles": ["white"],
        "expand": False,
        "title_justify": "left",
        "style": "black",
        "border_style": "black",
        "box": box.ROUNDED,
    }
    if headers:
        default.update(
            header_style="bold misty_rose1", box=box.SIMPLE_HEAVY, show_header=True
        )
    rows = kwargs.pop("rows", [])
    table = NewTable(*headers, **{**default, **kwargs})
    if rows:
        table.add_rows(rows)
    return table


def list_table(items: Iterable[Any], **kwargs: Any) -> NewTable:
    return new_table(rows=[[i] for i in items], **kwargs)


def _randint() -> int:
    return random.randint(50, 205)


def predictably_random_color(string: str) -> str:
    random.seed(string.strip())

    return f"#{_randint():02X}{_randint():02X}{_randint():02X}"


def _format_with_color(string: str, on: Optional[str] = None) -> str:
    color = f"b {predictably_random_color(string)}"
    if on:
        color += f" on {on}"
    return wrap(string, color)


def split_with_color(text: str) -> str:
    return " ".join(_format_with_color(str(x)) for x in sorted(SPLIT_PAT.split(text)))


def format_with_color(items: str | Sequence[str]) -> str:
    if isinstance(items, str):
        items = [items]

    return " ".join(_format_with_color(str(x)) for x in items)


def format_with_color_on_black(items: Union[str, Iterable[str]]) -> str:
    if not (isinstance(items, Iterable) and not isinstance(items, str)):
        items = sorted(SPLIT_PAT.split(str(items)))

    sep = wrap("a", "#000000 on #000000")
    return " ".join(
        sep + _format_with_color(str(item), on="#000000") + sep for item in items
    )


def fmt_pred_color(m: Match[str]) -> str:
    return f"{predictably_random_color(m.group(2))}]{m.group(2)}"


def simple_panel(content: RenderableType, **kwargs: Any) -> Panel:
    default: JSONDict = {
        "title_align": "left",
        "subtitle_align": "right",
        "box": box.SIMPLE,
        "expand": False,
        "border_style": "red",
    }
    if "title" in kwargs:
        kwargs["title"] = wrap(kwargs["title"], "b")
    if kwargs.pop("align", "") == "center":
        content = Align.center(content, vertical="middle")
    return Panel(content, **{**default, **kwargs})


def border_panel(content: RenderableType, **kwargs: Any) -> Panel:
    return simple_panel(
        content, **{"box": box.ROUNDED, "border_style": "dim", **kwargs}
    )


def md_panel(content: str, **kwargs: Any) -> Panel:
    return border_panel(
        Markdown(
            HTML_PARAGRAPH.sub("", content),
            inline_code_theme="nord-darker",
            code_theme="nord-darker",
            justify=kwargs.pop("justify", "left"),
        ),
        **kwargs,
    )


def new_tree(
    values: Iterable[RenderableType] = [], title: str = "", **kwargs: Any
) -> Tree:
    color = predictably_random_color(title or str(values))
    default: JSONDict = {"guide_style": color, "highlight": True}
    tree = Tree(title, **{**default, **kwargs})

    for val in values:
        tree.add(val)
    return tree


def get_country(code: str) -> str:
    return format_with_color(code)


def progress_bar(
    size: float, width: float, end: Optional[float] = None, inverse: bool = False
) -> Bar:
    if end is None:
        end = size
        size = width
        bgcolor = "black"
    else:
        bgcolor = "#252c3a"
    ratio = end / size if size else 1
    if inverse:
        ratio = 1 - ratio

    random.seed(str(width))

    def norm() -> int:
        return round(_randint() * ratio)

    color = f"#{norm():0>2X}{norm():0>2X}{norm():0>2X}"
    return Bar(
        size=size, begin=0, width=int(width), end=end, color=color, bgcolor=bgcolor
    )


def timestamp2datetime(timestamp: Union[str, int, float, None]) -> datetime:
    if isinstance(timestamp, str):
        timestamp = re.sub(r"[.]\d+", "", timestamp.strip("'"))
        formats = [
            "%Y-%m-%dT%H:%M:%SZ",
            "%Y%m%dT%H%M%SZ",
            "%Y-%m-%d %H:%M:%S",
            "%Y-%m-%dT%H:%M:%S%z",
        ]
        for fmt in formats:
            try:
                return datetime.strptime(timestamp, fmt)
            except ValueError:
                pass
    return datetime.fromtimestamp(int(float(timestamp or 0)), tz=timezone.utc)


def timestamp2timestr(timestamp: Union[str, int, float, None]) -> str:
    return timestamp2datetime(timestamp).strftime("%T")


def human_dt(timestamp: Union[int, str, float], acc: int = 1) -> str:
    try:
        datetime = timestamp2datetime(timestamp)
    except ValueError:
        return str(timestamp)

    diff = datetime.timestamp() - time.time()
    fmted = " ".join(islice(fmt_time(int(diff)), acc))

    return wrap(fmted, BOLD_RED if diff < 0 else BOLD_GREEN)


def diff_dt(timestamp: Union[int, str, float], acc: int = 2) -> str:
    try:
        datetime = timestamp2datetime(timestamp)
    except ValueError:
        return str(timestamp)

    diff = datetime.timestamp() - time.time()
    fmted = " ".join(islice(fmt_time(int(diff)), acc))

    strtime = datetime.strftime("%F" if abs(diff) > SECONDS_PER_DAY else "%T")

    return wrap(fmted, BOLD_RED if diff < 0 else BOLD_GREEN) + " " + strtime


def syntax(*args: Any, **kwargs: Any) -> Syntax:
    default = {
        "theme": "paraiso-dark",
        "background_color": "black",
        "word_wrap": True,
    }
    return Syntax(*args, **{**default, **kwargs})


def sql_syntax(sql_string: str) -> Syntax:
    return Syntax(
        sqlparse.format(
            sql_string,
            indent_columns=True,
            strip_whitespace=True,
            strip_comments=True,
            reindent=True,
            reindent_aligned=False,
        ),
        "sql",
        theme="gruvbox-dark",
        background_color="black",
        word_wrap=True,
    )


def diff_serialize(value: Any) -> str:
    if value is None:
        return "null"
    if value == "":
        return '""'
    return str(value)


@multimethod
def diff(before: str, after: str) -> Any:
    return make_difftext(before, after)


@diff.register
def _(before: Any, after: Any) -> Any:
    return diff(diff_serialize(before), diff_serialize(after))


@diff.register
def _(before: List[Any], after: List[Any]) -> Any:
    return list(starmap(diff, zip_longest(before, after)))


@diff.register
def _(before: List[str], after: List[str]) -> Any:
    before_set, after_set = set(before), set(after)
    common = before_set & after_set
    common_list = list(common)
    return [
        *list(starmap(diff, zip(common_list, common_list))),
        *list(
            starmap(
                diff,
                zip_longest(list(before_set - common), list(after_set - common)),
            )
        ),
    ]


@diff.register
def _(before: dict, after: dict) -> Any:
    data = {}
    keys = sorted(before.keys() | after.keys())
    for key in keys:
        if key not in before:
            data[wrap(key, BOLD_GREEN)] = wrap(diff_serialize(after[key]), BOLD_GREEN)
        elif key not in after:
            data[wrap(key, BOLD_RED)] = wrap(diff_serialize(before[key]), BOLD_RED)
        else:
            data[key] = diff(before[key], after[key])

    return data


def pretty_diff(before: Any, after: Any) -> Text:
    result = diff(before, after)
    if not isinstance(result, str):
        result = (
            pformat(result, indent=2, width=300, sort_dicts=False)
            .replace("'", "")
            .replace("\\\\", "\\")
        )

    return Text.from_markup(result)
